# -*- coding: utf-8 -*-
#


class SurfaceLoop(object):
    _ID = 0
    dimension = 2

    def __init__(self, surfaces):
        self.surfaces = surfaces

        self.id = 'sl{}'.format(SurfaceLoop._ID)
        SurfaceLoop._ID += 1

        self.code = '\n'.join([
            '{} = news;'.format(self.id),
            'Surface Loop({}) = {{{}}};'.format(
                self.id, ','.join([s.id for s in surfaces])
            )])
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from itertools import permutations

import pygmsh

from helpers import compute_volume


def test(lcar=1.):
    lbw = [2, 3, 5]
    geom = pygmsh.built_in.Geometry()
    points = [geom.add_point([x, 0., 0.], lcar) for x in [0., lbw[0]]]
    line = geom.add_line(*points)

    _, rectangle, _ = geom.extrude(line,
                                   translation_axis=[0., lbw[1], 0.],
                                   num_layers=lbw[1],
                                   recombine=True)
    geom.extrude(rectangle,
                 translation_axis=[0., 0., lbw[2]],
                 num_layers=lbw[2],
                 recombine=True)

    # compute_volume only supports 3D for tetras, but does return
    # surface area for quads

    ref = sum(l * w for l, w in permutations(lbw, 2))  # surface area
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('hex.vtu', *test())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Creates a mesh for a square with a round hole.
'''
import pygmsh

from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    circle = geom.add_circle(
        x0=[0.5, 0.5, 0.0],
        radius=0.25,
        lcar=0.1,
        num_sections=4,
        make_surface=False
        )

    geom.add_rectangle(
        0.0, 1.0,
        0.0, 1.0,
        0.0,
        lcar=0.1,
        holes=[circle.line_loop]
        )

    ref = 0.8086582838174551
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('rectangle_with_hole.vtu', *test())



# -*- coding: utf-8 -*-
#
from .volume_base import VolumeBase


class Cylinder(VolumeBase):
    def __init__(self, x0, axis, radius, angle=None, char_length=None):
        super(Cylinder, self).__init__()

        assert len(x0) == 3
        assert len(axis) == 3

        self.x0 = x0
        self.axis = axis
        self.radius = radius
        self.angle = angle
        self.char_length = char_length

        args = list(x0) + list(axis) + [radius]
        if angle is not None:
            args.append(angle)
        args = ', '.join(['{}'.format(arg) for arg in args])

        code = [
            '{} = newv;'.format(self.id),
            'Cylinder({}) = {{{}}};'.format(self.id, args)
            ]

        if self.char_length:
            code.extend([
                'pts_{}[] = PointsOf{{Volume{{{}}};}};'.format(
                    self.id, self.id
                    ),
                'Characteristic Length{{pts_{}[]}} = {};'.format(
                    self.id, char_length
                    ),
                ])

        self.code = '\n'.join(code)
        return



# -*- coding: utf-8 -*-
#
from .volume_base import VolumeBase


class Ball(VolumeBase):
    def __init__(
            self, center, radius, x0=None, x1=None, alpha=None,
            char_length=None
            ):
        '''Generate a solid ball.

        Parameters
        ----------
        center: list of 3 floats
           Center of the ball.

        radius: float
           Radius of the ball.

        x0: float (optional)
           If specified and `x0 > -1`, the ball is cut off at `x0*radius`
           parallel to the y-z plane.

        x1: float (optional)
           If specified and `x1 < +1`, the ball is cut off at `x1*radius`
           parallel to the y-z plane.

        alpha: float (optional)
           If specified and `alpha < 2*pi`, the points between `alpha` and
           `2*pi` w.r.t. to the x-y plane are not part of the object.

        char_length: float (optional)
           If specified, sets the `Characteristic Length` property.
        '''
        super(Ball, self).__init__()

        self.center = center
        self.radius = radius
        self.char_length = char_length

        args = list(center) + [radius]
        if x0 is not None:
            args.append(x0)
            if x1 is not None:
                args.append(x1)
                if alpha is not None:
                    args.append(alpha)
        args = ', '.join(['{}'.format(arg) for arg in args])

        self.code = '\n'.join([
            '{} = newv;'.format(self.id),
            'Sphere({}) = {{{}}};'.format(self.id, args)
            ] + self.char_length_code(char_length)
            )
        return



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pygmsh

from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    geom.add_circle(
        [0.0, 0.0, 0.0],
        1.0,
        0.1,
        num_sections=4,
        # If compound==False, the section borders have to be points of the
        # discretization. If using a compound circle, they don't; gmsh can
        # choose by itself where to point the circle points.
        compound=True
        )

    ref = 3.1363871677682247
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('circle.vtk', *test())



# -*- coding: utf-8 -*-
#
from .line_base import LineBase
from .point import Point


class Bspline(LineBase):
    def __init__(self, control_points):
        super(Bspline, self).__init__()

        for c in control_points:
            assert isinstance(c, Point)
        assert len(control_points) > 1

        self.control_points = control_points

        self.code = '\n'.join([
            '{} = newl;'.format(self.id),
            'BSpline({}) = {{{}}};'.format(
                self.id, ', '.join([c.id for c in self.control_points])
            )])
        return



# -*- coding: utf-8 -*-
#
from __future__ import print_function

import os
import subprocess
import tempfile

import numpy

import meshio


def rotation_matrix(u, theta):
    '''Return matrix that implements the rotation around the vector :math:`u`
    by the angle :math:`\\theta`, cf.
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle.

    :param u: rotation vector
    :param theta: rotation angle
    '''
    assert numpy.isclose(numpy.inner(u, u), 1.), \
        'the rotation axis must be unitary'

    # Cross-product matrix.
    cpm = numpy.array([
        [0.0, -u[2], u[1]],
        [u[2], 0.0, -u[0]],
        [-u[1], u[0], 0.0]
        ])
    c = numpy.cos(theta)
    s = numpy.sin(theta)
    R = numpy.eye(3) * c \
        + s * cpm \
        + (1.0 - c) * numpy.outer(u, u)
    return R


def _is_string(obj):
    try:
        # Python 2
        return isinstance(obj, basestring)
    except NameError:
        # Python 3
        return isinstance(obj, str)


def _is_flat(X, tol=1.0e-15):
    '''Checks if all points X sit in a plane.
    '''
    # find three points that don't sit on a line
    found = False
    for x2 in X:
        orth = numpy.cross(X[1] - X[0], x2 - X[0])
        orth_dot_orth = numpy.dot(orth, orth)
        if orth_dot_orth > tol:
            found = True
            break
    if not found:
        # All points even sit on a line
        return True
    norm_orth = numpy.sqrt(orth_dot_orth)
    norm_x_min_x0 = numpy.sqrt(numpy.einsum('ij, ij->i', X - X[0], X - X[0]))
    return (
        abs(numpy.dot(X - X[0], orth)) < tol * (1.0 + norm_orth*norm_x_min_x0)
        ).all()


def _get_gmsh_exe():
    macos_gmsh_location = '/Applications/Gmsh.app/Contents/MacOS/gmsh'
    if os.path.isfile(macos_gmsh_location):
        gmsh_executable = macos_gmsh_location
    else:
        gmsh_executable = 'gmsh'
    return gmsh_executable


def get_gmsh_major_version(gmsh_exe=_get_gmsh_exe()):
    out = subprocess.check_output(
            [gmsh_exe, '--version'],
            stderr=subprocess.STDOUT
            ).strip().decode('utf8')
    ex = out.split('.')
    return int(ex[0])


# pylint: disable=too-many-branches
def generate_mesh(
        geo_object,
        optimize=True,
        verbose=True,
        dim=3,
        prune_vertices=True,
        gmsh_path=None,
        geom_order=1,
        # for debugging purposes:
        geo_filename=None
        ):
    preserve_geo = geo_filename is not None
    if geo_filename is None:
        with tempfile.NamedTemporaryFile(suffix='.geo') as f:
            geo_filename = f.name

    with open(geo_filename, 'w') as f:
        f.write(geo_object.get_code())

    with tempfile.NamedTemporaryFile(suffix='.msh') as handle:
        msh_filename = handle.name

    gmsh_executable = gmsh_path if gmsh_path is not None else _get_gmsh_exe()

    cmd = [
        gmsh_executable,
        '-{}'.format(dim), '-bin', geo_filename, '-o', msh_filename
        ]

    gmsh_major_version = get_gmsh_major_version(gmsh_executable)
    if gmsh_major_version < 3 and optimize:
        cmd += ['-optimize']

    assert geom_order > 0
    if geom_order > 1:
        cmd += ['-order', str(geom_order)]

    # https://stackoverflow.com/a/803421/353337
    p = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
    if verbose:
        while True:
            line = p.stdout.readline()
            if not line:
                break
            print(line.decode('utf-8'), end='')

    p.communicate()
    assert p.returncode == 0, \
        'Gmsh exited with error (return code {}).'.format(p.returncode)

    X, cells, pt_data, cell_data, field_data = meshio.read(msh_filename)

    # Lloyd smoothing
    if not _is_flat(X) or 'triangle' not in cells:
        if verbose:
            print(
                'Not performing Lloyd smoothing '
                '(only works for flat triangular meshes).'
                )
        return X, cells, pt_data, cell_data, field_data

    if prune_vertices:
        # Make sure to include only those vertices which belong to a triangle.
        uvertices, uidx = numpy.unique(cells['triangle'], return_inverse=True)
        cells = {'triangle': uidx.reshape(cells['triangle'].shape)}
        cell_data = {'triangle': cell_data['triangle']}
        X = X[uvertices]
        for key in pt_data:
            pt_data[key] = pt_data[key][uvertices]

    # clean up
    os.remove(msh_filename)
    if preserve_geo:
        print('\ngeo file: {}'.format(geo_filename))
    else:
        os.remove(geo_filename)
    return X, cells, pt_data, cell_data, field_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pygmsh


def test(lcar=1.):
    geom = pygmsh.built_in.Geometry()
    poly = geom.add_polygon([
        [0., 0., 0.],
        [1., 0., 0.],
        [1., 1., 0.],
        [0., 1., 0.]],
        lcar
        )

    geom.set_transfinite_surface(poly.surface, size=[11, 9])

    points, cells, _, _, _ = pygmsh.generate_mesh(
        geom,
        geo_filename='transfinite.geo'
        )
    assert len(cells['triangle']) == 10*8*2
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('transfinite.vtu', *test())



# -*- coding: utf-8 -*-
#
from .. import built_in

class SurfaceBase(built_in.surface_base.SurfaceBase):
    _ID = 0
    dimension = 2

    def __init__(self, is_list=False, id0=None):
        super(SurfaceBase, self).__init__(id0=id0)

        self.is_list = is_list
        if is_list:
            self.id += '[]'
        return

    def char_length_code(self, char_length):
        if char_length is None:
            return []

        return [
            'pts_{}[] = PointsOf{{Surface{{{}}};}};'.format(
                self.id, self.id
                ),
            'Characteristic Length{{pts_{}[]}} = {};'.format(
                self.id, char_length
                ),
            ]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pygmsh

from helpers import compute_volume



def test(radius=1.):
    geom = pygmsh.built_in.Geometry()

    R = [pygmsh.rotation_matrix(np.eye(1, 3, d)[0], theta)
         for d, theta in enumerate(np.pi / np.array([2., 3., 5]))]

    geom.add_circle(
        [7., 11., 13.],
        radius,
        .1,
        R[0] @ R[1] @ R[2])

    ref = np.pi * radius ** 2
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert np.isclose(compute_volume(points, cells), ref, rtol=1e-2)
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('circle_transformed.vtk', *test())



# -*- coding: utf-8 -*-
#
from .surface_base import SurfaceBase


class Disk(SurfaceBase):
    def __init__(self, x0, radius0, radius1=None, char_length=None):
        super(Disk, self).__init__()

        assert len(x0) == 3
        if radius1 is not None:
            assert radius0 >= radius1

        self.x0 = x0
        self.radius0 = radius0
        self.radius1 = radius1
        self.char_length = char_length

        args = list(x0) + [radius0]
        if radius1 is not None:
            args.append(radius1)

        args = ', '.join(['{}'.format(arg) for arg in args])

        self.code = '\n'.join([
            '{} = news;'.format(self.id),
            'Disk({}) = {{{}}};'.format(self.id, args)
            ] + self.char_length_code(char_length)
            )
        return



# -*- coding: utf-8 -*-
#


class CompoundVolume(object):
    _ID = 0

    def __init__(self, volumes):
        self.volumes = volumes

        self.id = 'cv{}'.format(CompoundVolume._ID)
        CompoundVolume._ID += 1

        self.code = '\n'.join([
            '{} = newv;'.format(self.id),
            'Compound Volume({}) = {{{}}};'.format(
                self.id, ','.join([v.id for v in volumes])
            )])
        return



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pytest

import pygmsh

from helpers import compute_volume


@pytest.mark.skipif(
    pygmsh.get_gmsh_major_version() < 3,
    reason='requires Gmsh >= 3'
    )
def test():
    geom = pygmsh.opencascade.Geometry(
        characteristic_length_min=2.0,
        characteristic_length_max=2.0,
        )

    rect1 = geom.add_rectangle(
        [10.0, 0.0, 0.0], 20.0, 40.0,
        corner_radius=5.0,
        )
    rect2 = geom.add_rectangle(
        [0.0, 10.0, 0.0], 40.0, 20.0,
        corner_radius=5.0
        )
    disk1 = geom.add_disk([14.5, 35.0, 0.0], 1.85)
    disk2 = geom.add_disk([25.5, 5.0, 0.0], 1.85)

    rect3 = geom.add_rectangle([10.0, 30.0, 0.0], 10.0, 1.0)
    rect4 = geom.add_rectangle([20.0, 9.0, 0.0], 10.0, 1.0)

    r1 = geom.add_rectangle([9.0, 0.0, 0.0], 21.0, 20.5, corner_radius=8.0)
    r2 = geom.add_rectangle([10.0, 00.0, 0.0], 20.0, 19.5, corner_radius=7.0)
    diff1 = geom.boolean_difference([r1], [r2])
    r22 = geom.add_rectangle([9.0, 10.0, 0.0], 11.0, 11.0)
    inter1 = geom.boolean_intersection([diff1, r22])

    r3 = geom.add_rectangle([10.0, 19.5, 0.0], 21.0, 21.0, corner_radius=8.0)
    r4 = geom.add_rectangle([10.0, 20.5, 0.0], 20.0, 20.0, corner_radius=7.0)
    diff2 = geom.boolean_difference([r3], [r4])
    r33 = geom.add_rectangle([20.0, 19.0, 0.0], 11.0, 11.0)
    inter2 = geom.boolean_intersection([diff2, r33])

    geom.boolean_difference(
        [rect1, rect2],
        [disk1, disk2, rect3, rect4, inter1, inter2]
        )

    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    ref = 1082.4470502181903
    assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    from helpers import plot
    plot('logo.png', *test())
    # import meshio
    # meshio.write('logo.vtu', *test())



# -*- coding: utf-8 -*-
#
from .surface_base import SurfaceBase


class CompoundSurface(SurfaceBase):
    def __init__(self, surfaces):
        super(CompoundSurface, self).__init__()
        self.num_edges = sum(s.num_edges for s in surfaces)

        self.surfaces = surfaces

        self.code = '\n'.join([
            '{} = news;'.format(self.id),
            'Compound Surface({}) = {{{}}};'.format(
                self.id, ','.join([s.id for s in surfaces])
            )])
        return



# -*- coding: utf-8 -*-
#
from .line_loop import LineLoop


class Surface(object):
    _ID = 0
    num_edges = 0
    dimension = 2

    def __init__(self, line_loop, api_level=2):
        assert isinstance(line_loop, LineLoop)

        self.line_loop = line_loop

        self.id = 'rs{}'.format(Surface._ID)
        Surface._ID += 1

        # `Ruled Surface` was deprecated in Gmsh 3 in favor of `Surface`.
        name = 'Surface' if api_level > 2 else 'Ruled Surface'

        self.code = '\n'.join([
            '{} = news;'.format(self.id),
            '{}({}) = {{{}}};'.format(name, self.id, self.line_loop.id)
            ])
        self.num_edges = len(line_loop)
        return



# -*- coding: utf-8 -*-
#
from .line_base import LineBase


class CompoundLine(LineBase):
    def __init__(self, lines):
        super(CompoundLine, self).__init__()

        self.lines = lines

        self.code = '\n'.join([
            '{} = newl;'.format(self.id),
            'Compound Line({}) = {{{}}};'.format(
                self.id, ','.join([l.id for l in self.lines])
            )])
        return



# -*- coding: utf-8 -*-
#
from .volume_base import VolumeBase


class Torus(VolumeBase):
    def __init__(self, center, radius0, radius1, alpha=None, char_length=None):
        super(Torus, self).__init__()

        assert len(center) == 3

        self.center = center
        self.radius0 = radius0
        self.radius1 = radius1
        self.alpha = alpha
        self.char_length = char_length

        args = list(center) + [radius0] + [radius1]
        if alpha is not None:
            args.append(alpha)
        args = ', '.join(['{}'.format(arg) for arg in args])

        self.code = '\n'.join([
            '{} = newv;'.format(self.id),
            'Torus({}) = {{{}}};'.format(self.id, args)
            ] + self.char_length_code(char_length)
            )
        return



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
import numpy as np
import pygmsh


def test():
    # Characteristic length
    lcar = 1e-1

    # Coordinates of lower-left and upper-right vertices of a square domain
    xmin = 0.0
    xmax = 5.0
    ymin = 0.0
    ymax = 5.0

    # Vertices of a square hole
    squareHoleCoordinates = np.array([
        [1, 1, 0],
        [4, 1, 0],
        [4, 4, 0],
        [1, 4, 0]
        ])

    # Create geometric object
    geom = pygmsh.built_in.Geometry()

    # Create square hole
    squareHole = geom.add_polygon(
        squareHoleCoordinates, lcar,
        make_surface=False
        )

    # Create square domain with square hole
    geom.add_rectangle(
        xmin, xmax, ymin, ymax, 0.0, lcar,
        holes=[squareHole.line_loop]
        )

    points, cells, _, _, _ = pygmsh.generate_mesh(geom, geom_order=2)
    # TODO support for volumes of triangle6
    # ref = 16.0
    # from helpers import compute_volume
    # assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('hole_in_square.vtu', *test())



# -*- coding: utf-8 -*-
#
from .surface_base import SurfaceBase


class Rectangle(SurfaceBase):
    def __init__(self, x0, a, b, corner_radius=None, char_length=None):
        super(Rectangle, self).__init__()

        assert len(x0) == 3

        self.x0 = x0
        self.a = a
        self.b = b
        self.corner_radius = corner_radius
        self.char_length = char_length

        args = list(x0) + [a, b]
        if corner_radius is not None:
            args.append(corner_radius)

        args = ', '.join(['{}'.format(arg) for arg in args])

        self.code = '\n'.join([
            '{} = news;'.format(self.id),
            'Rectangle({}) = {{{}}};'.format(self.id, args)
            ] + self.char_length_code(char_length)
            )
        return



# -*- coding: utf-8 -*-
#


class Point(object):
    _POINT_ID = 0

    def __init__(self, x, lcar=None):
        self.x = x
        self.lcar = lcar

        self.id = 'p{}'.format(Point._POINT_ID)
        Point._POINT_ID += 1

        # Points are always 3D in gmsh
        if lcar is not None:
            self.code = '\n'.join([
                '{} = newp;'.format(self.id),
                'Point({}) = {{{!r}, {!r}, {!r}, {!r}}};'.format(
                    self.id, x[0], x[1], x[2], lcar
                )])
        else:
            self.code = '\n'.join([
                '{} = newp;'.format(self.id),
                'Point({}) = {{{!r}, {!r}, {!r}}};'.format(
                    self.id, x[0], x[1], x[2]
                )])
        return



# -*- coding: utf-8 -*-
#
from .line_base import LineBase
from .point import Point


class Spline(LineBase):
    def __init__(self, points):
        super(Spline, self).__init__()

        for c in points:
            assert isinstance(c, Point)
        assert len(points) > 1

        self.points = points

        self.code = '\n'.join([
            '{} = newl;'.format(self.id),
            'Spline({}) = {{{}}};'.format(
                self.id, ', '.join([c.id for c in self.points])
            )])
        return



# -*- coding: utf-8 -*-
#


class LineLoop(object):
    _ID = 0
    dimension = 1

    def __init__(self, lines):
        self.lines = lines

        self.id = 'll{}'.format(LineLoop._ID)
        LineLoop._ID += 1

        self.code = '\n'.join([
            '{} = newll;'.format(self.id),
            'Line Loop({}) = {{{}}};'.format(
                self.id, ', '.join([l.id for l in lines])
            )])
        return

    def __len__(self):
        return len(self.lines)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pygmsh

from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    geom.add_rectangle(
        0.0, 1.0,
        0.0, 1.0,
        0.0,
        0.1
        )

    ref = 1.0
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('rectangle.vtu', *test())



# -*- coding: utf-8 -*-
#


class VolumeBase(object):
    _ID = 0
    dimension = 3

    def __init__(self, id0=None):
        if id0:
            self.id = id0
        else:
            self.id = 'vol{}'.format(VolumeBase._ID)
            VolumeBase._ID += 1
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pygmsh

from helpers import compute_volume


def test(irad=0.05,
         orad=0.6):
    '''Torus, rotated in space.
    '''
    geom = pygmsh.built_in.Geometry()

    R = pygmsh.rotation_matrix([1., 0., 0.], np.pi / 2)
    geom.add_torus(
        irad=irad, orad=orad, lcar=0.03,
        x0=[0.0, 0.0, -1.0],
        R=R
        )

    ref = 2 * np.pi ** 2 * orad * irad ** 2
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert np.isclose(compute_volume(points, cells), ref,
                      rtol=5e-2)
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('torus.vtu', *test())



# -*- coding: utf-8 -*-
#
from .volume_base import VolumeBase


class Wedge(VolumeBase):
    def __init__(self, x0, extents, top_extent=None, char_length=None):
        super(Wedge, self).__init__()

        self.x0 = x0
        self.extents = extents
        self.top_extent = top_extent
        self.char_length = char_length

        args = list(x0) + list(extents)
        if top_extent is not None:
            args.append(top_extent)
        args = ', '.join(['{}'.format(arg) for arg in args])

        self.code = '\n'.join([
            '{} = newv;'.format(self.id),
            'Wedge({}) = {{{}}};'.format(self.id, args)
            ] + self.char_length_code(char_length)
            )
        return



# -*- coding: utf-8 -*-
#

from .volume_base import VolumeBase

class Volume(VolumeBase):
    def __init__(self, surface_loop, holes=None):
        super(Volume, self).__init__()

        if holes is None:
            holes = []

        self.surface_loop = surface_loop
        self.holes = holes

        surface_loops = [surface_loop] + holes

        self.code = '\n'.join([
            '{} = newv;'.format(self.id),
            'Volume({}) = {{{}}};'.format(
                self.id, ', '.join([s.id for s in surface_loops])
            )])
        return



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pytest

import pygmsh

from helpers import compute_volume


@pytest.mark.skipif(
    pygmsh.get_gmsh_major_version() < 3,
    reason='requires Gmsh >= 3'
    )
def test():
    geom = pygmsh.opencascade.Geometry()

    geom.add_box([0.0, 0.0, 0.0], [1, 2, 3], char_length=0.1)

    ref = 6
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('opencascade_box.vtu', *test())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Creates a mesh for an ellipsoid.
'''
import pygmsh

from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()
    geom.add_ellipsoid(
        [0.0, 0.0, 0.0],
        [1.0, 0.5, 0.75],
        0.05
        )

    ref = 1.5676038497587947
    points, cells, _, _, _ = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return points, cells


if __name__ == '__main__':
    import meshio
    meshio.write('ellipsoid.vtu', *test())
